##################################################
### script to plot the future extinction proba ###
###          (output from MIDASPOM)            ###
##################################################

print("Running Rscript to plot the future probability of extinction")

#recover arguments of script
library("optparse")

option_list = list(
  make_option(c("-o", "--out"), type="character", default="out", help="output pdf name [default= %default]", metavar="character"),
  make_option(c("-i", "--input"), type="character", default="pext.txt", help="input file with number of simulations with extinction [default= %default]", metavar="character"),
  make_option(c("-t", "--tmin"), type="numeric", default=0, help="minimum t [default= %default]", metavar="number"),
  make_option(c("-T", "--tmax"), type="numeric", default=1, help="maximum T [default= %default]", metavar="number"),
  make_option(c("-n", "--nrep"), type="numeric", default=1, help="number of simulations [default= %default]", metavar="number")
); 

opt_parser = OptionParser(option_list=option_list);
opt = parse_args(opt_parser);

#parameters
prextfile = opt$input #file with probability of extinction as a function of time
outfile   = opt$out #output file name
tmin = as.numeric(opt$tmin)
tmax = as.numeric(opt$tmax)
nrep = as.numeric(opt$nrep)

#load joint posterior density
pext  = scan(prextfile)

#plot
t = tmin:tmax
pdf(outfile,h=3.5,w=3.5)
par(mfrow=c(1,1),family="Times",las=1)
plot(t, pext/nrep ,type="l",ylim=c(0,1),xlab="year",ylab="Probability of extinction",lwd=2)
dev.off()

print(paste("plotted to file",outfile))
