#################################################
### script to plot the imputed occupancies    ###
###          (output from MIDASPOM)           ###
#################################################

print("Running Rscript to plot the imputed occupancies")

#recover arguments of script
library("optparse")

option_list = list(
  make_option(c("-o", "--out"), type="character", default="out", help="output pdf name [default= %default]", metavar="character"),
  make_option(c("-i", "--inocc"), type="character", default="occupancy.txt", help="input file with proportion of occupied patches [default= %default]", metavar="character"),
  make_option(c("-I", "--inpatch"), type="character", default="patches.txt", help="input file with occupancy probability for each patch [default= %default]", metavar="character"),
  make_option(c("-t", "--tmin"), type="numeric", default=0, help="minimum t [default= %default]", metavar="number"),
  make_option(c("-T", "--tmax"), type="numeric", default=1, help="maximum T [default= %default]", metavar="number")
); 

opt_parser = OptionParser(option_list=option_list);
opt = parse_args(opt_parser);

#parameters
inocc    = opt$inocc
inpatch  = opt$inpatch
outfile = opt$out # args[2] #file with posterior distribution
tmin  = as.numeric( opt$tmin) #args[3])
tmax  = as.numeric( opt$tmax) #args[4])
print("Arguments read")

#read files
impoc = read.table(inocc)
imp   = read.table(inpatch)
n = ncol(imp)

print(tmin:tmax)
print(impoc)

pdf(opt$out,h=3.5,w=4*2)
coco = rgb(1,seq(1,0,-0.1),seq(1,0,-0.1),1)
par(mfrow=c(1,2),family="Times",las=1)
image(tmin:tmax,0:(n+1),as.matrix(impoc),asp=1/1.1 ,col=coco,xlab="Time (years)",ylab="Occupancy (# of patches)",useRaster = F,axes=F,breaks = seq(0,1,length.out = 12),xlim=c(tmin-0.5,tmax+0.5))
rect(tmin-0.5,0,tmax+0.5,n+1 )
axis(1,seq(tmin,tmax,1),pos=0)
axis(2,seq(0,n,1),at=seq(0,n,1)+0.5,pos=tmin-0.5)
image(tmin:tmax,1:n,as.matrix(imp),asp=1/1.1 ,col=coco,xlab="Time (years)",ylab="Segment",useRaster = F,axes=F,breaks = seq(0,1,length.out = 12),xlim=c(tmin-0.5,tmax+0.5))
rect(tmin-0.5,0.5,tmax+0.5,n+0.5 )
axis(1,seq(tmin,tmax,1),pos=0.5)
axis(2,seq(1,n,1),at=seq(1,n,1),pos=tmin-0.5)
dev.off()
