#################################################
### script to plot the posterior distribution ###
###          (output from MIDASPOM)           ###
#################################################

print("Running Rscript to plot the posterior density")

#recover arguments of script
library("optparse")

option_list = list(
  make_option(c("-o", "--out"), type="character", default="out", help="output pdf name [default= %default]", metavar="character"),
  make_option(c("-O", "--outpost"), type="character", default="posterior.txt", help="output posterior file name [default= %default]", metavar="character"),
  make_option(c("-y", "--ylim"), type="numeric", default=0.5, help="limit of y-axis [default= %default]", metavar="number"),
  make_option(c("-z", "--zlim"), type="numeric", default=1, help="limit of z-axis [default= %default]", metavar="number"),
  make_option(c("-p", "--pattern"), type="character", default="dieoff", help="pattern for input file names [default= %default]", metavar="character"),
  make_option(c("-e", "--emin"), type="numeric", default=0, help="minimum e [default= %default]", metavar="number"),
  make_option(c("-E", "--emax"), type="numeric", default=1, help="maximum e [default= %default]", metavar="number"),
  make_option(c("-c", "--cmin"), type="numeric", default=0, help="minimum c [default= %default]", metavar="number"),
  make_option(c("-C", "--cmax"), type="numeric", default=1, help="maximum c [default= %default]", metavar="number"),
  make_option(c("-q", "--pmin"), type="numeric", default=0, help="minimum p [default= %default]", metavar="number"),
  make_option(c("-Q", "--pmax"), type="numeric", default=1, help="maximum p [default= %default]", metavar="number"),
  make_option(c("-s", "--step"), type="numeric", default=0.01, help="Step e and c [default= %default]", metavar="number"),
  make_option(c("-t", "--stepp"), type="numeric", default=0.05, help="Step p [default= %default]", metavar="number")
); 

opt_parser = OptionParser(option_list=option_list);
opt = parse_args(opt_parser);

#parameters
postfilespattern = opt$pattern #file with posterior distribution of parameters under in habitat loss hypothesis
outfile          = opt$out # args[2] #file with posterior distribution
emin  = as.numeric( opt$emin) #args[3])
emax  = as.numeric( opt$emax) #args[4])
cmin  = as.numeric( opt$cmin) #args[5])
cmax  = as.numeric( opt$cmax) #args[6])
pmin  = as.numeric( opt$pmin) #args[5])
pmax  = as.numeric( opt$pmax) #args[6])
step  = as.numeric( opt$step)
stepp  = as.numeric( opt$stepp)
print("Arguments read")

el = seq(emin,emax,step)
cl = seq(cmin,cmax,step)
pl = seq(pmin,pmax,stepp)

#print(pl)

#load joint posterior density
postfiles  = grep(postfilespattern, list.files(),value = T,perl = T)
jpostl = lapply(postfiles, function(x) as.matrix(read.table(x)) ) #matrix(scan(postfiledieoff),1,stepK,byrow=T)
jpostl = lapply( jpostl, function(x) exp(x) )

jpost = Reduce('+',jpostl)
jpost = jpost/sum(jpost)/step/step
#print(sum(jpost)*step*step)

#point estimates
ML   =  which(jpost==max(jpost),arr.ind = T)
nstepe = (emax-emin)/step +1
nstepc = (cmax-cmin)/step +1
nstepp = (pmax-pmin)/stepp +1
eest = el[ML[1]]
cest = cl[ML[2]]


#compute marginal posterior densities
epost = rowMeans(jpost)
cpost = colMeans(jpost)
ppost = sapply(jpostl,sum)
ppost = ppost/sum(ppost)/stepp

pest = pl[which.max(ppost)]

#print( which.max(ppost) )

qel = which((cumsum(epost/sum(epost))>=(0.025) )&(cumsum(epost/sum(epost))<(0.975)))
qcl = which((cumsum(cpost/sum(cpost))>=(0.025) )&(cumsum(cpost/sum(cpost))<(0.975)))
qpl = which((cumsum(ppost/sum(ppost))>=(0.025) )&(cumsum(ppost/sum(ppost))<(0.975)))

#print(ppost)
print(c(pest,eest,cest))
#plot
coco = rgb(1,seq(1,0,-0.1),seq(1,0,-0.1),1)

pdf(outfile,h=1*2.5,w=4*2.5)
par(mfrow=c(1,4),family="Times",las=1)
plot( -1, -1,type="l",xlim=c(pmin,pmax),ylim=c(0,8), xlab="detection probability",ylab="density",col=2,lwd=3,las=1,main=" ")
polygon(c(pmin,pmin,pmax,pmax),c(0,1/(pmax-pmin),1/(pmax-pmin),0),lwd=2,col=rgb(0,0,0,0.3),border=NA)
polygon(c(pl,rev(pl)), c(ppost,rep(0,nstepp) ),col=rgb(1,0,0,0.3),border=NA ) 
polygon(c(pl[qpl],rev(pl[qpl])), c(ppost[qpl],rep(0,nstepp)[qpl] ),col=rgb(1,0,0,0.5),border=NA ) 
abline(v=pl[qpl[1]],col=2,lwd=3,lty=2)
abline(v=pl[rev(qpl)[1]],col=2,lwd=3,lty=2)

plot( -1, -1,type="l",xlim=c(emin,emax),ylim=c(0,8), xlab="extinction parameter",ylab="density",col=2,lwd=3,las=1,main=" ")
polygon(c(emin,emin,emax,emax),c(0,1/(emax-emin),1/(emax-emin),0),lwd=2,col=rgb(0,0,0,0.3),border=NA)
polygon(c(el,rev(el)), c(epost,rep(0,nstepe) ),col=rgb(1,0,0,0.3),border=NA ) 
polygon(c(el[qel],rev(el[qel])), c(epost[qel],rep(0,nstepe)[qel] ),col=rgb(1,0,0,0.5),border=NA ) 
abline(v=el[qel[1]],col=2,lwd=3,lty=2)
abline(v=el[rev(qel)[1]],col=2,lwd=3,lty=2)

plot( -1, -1,type="l",xlim=c(cmin,cmax),ylim=c(0,8), xlab="colonization parameter",ylab="density",col=2,lwd=3,las=1)
polygon(c(cmin,cmin,cmax,cmax),c(0,1/(cmax-cmin),1/(cmax-cmin),0),lwd=2,col=rgb(0,0,0,0.3),border=NA)
polygon(c(cl,rev(cl)), c(cpost,rep(0,nstepc) ),col=rgb(1,0,0,0.5),border=NA ) 
polygon(c(cl[qcl],rev(cl[qcl])), c(cpost[qcl],rep(0,nstepc)[qcl] ),col=rgb(1,0,0,0.5),border=NA ) 
abline(v=cl[qcl[1]],col=2,lwd=3,lty=2)
abline(v=cl[rev(qcl)[1]],col=2,lwd=3,lty=2)

image(el,cl,jpost,col=coco,xlab="extinction parameter",ylab="colonization parameter",useRaster = T)
abline(v=eest,lty=2,col=2)
abline(h=cest,lty=2,col=2)
mtext(eest,side=1,at=eest,col=2,line=0.5)
mtext(cest,side=2,at=cest,col=2,line=0.5)
dev.off()

write.table(jpost,file=opt$outpost,quote=F,row.names=F,col.names=F,sep="\t")

print(paste("plotted to file",outfile))
